/***********************************************************************
 *                                                                      *
 *               This software is part of the ast package               *
 *          Copyright (c) 1985-2011 AT&T Intellectual Property          *
 *                      and is licensed under the                       *
 *                 Eclipse Public License, Version 1.0                  *
 *                    by AT&T Intellectual Property                     *
 *                                                                      *
 *                A copy of the License is available at                 *
 *          http://www.eclipse.org/org/documents/epl-v10.html           *
 *         (with md5 checksum b35adb5213ca9657e911e9befb180842)         *
 *                                                                      *
 *              Information and Software Systems Research               *
 *                            AT&T Research                             *
 *                           Florham Park NJ                            *
 *                                                                      *
 *               Glenn Fowler <glenn.s.fowler@gmail.com>                *
 *                    David Korn <dgkorn@gmail.com>                     *
 *                     Phong Vo <phongvo@gmail.com>                     *
 *                                                                      *
 ***********************************************************************/
/*
 * Glenn Fowler
 * AT&T Research
 *
 * generic error discipline function
 */
#include "config_ast.h"  // IWYU pragma: keep

#include <stdarg.h>

#include "ast_errorf.h"
#include "error.h"

int errorf(void *handle, void *discipline, int level, ...) {
    va_list ap;

    va_start(ap, level);
    errorv((discipline && handle) ? *((char **)handle) : (char *)handle,
           (discipline || level < 0) ? level : (level | ERROR_LIBRARY), ap);
    va_end(ap);
    return 0;
}
